# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

"""
This package contains standard serializers that can work generally for all
pymatgen objects.
"""
