# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

"""
The symmetry package implements symmetry tools, e.g., spacegroup determination,
etc.
"""
