# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

"""
The phasediagram package implements the analysis tools to perform phase
stability analyses, including the constructing of phase diagrams, determination
of decomposition products, etc. The package is designed to be fairly modular
and standalone.
"""

__author__ = "Shyue"
__date__ = "Mar 28 2013"
